﻿using Squid;
using System;
using System.Collections;
using System.Collections.Generic;
using System.ComponentModel;

public class GridViewItemCollection : ActiveList<GridViewItem> { }

[Toolbox]
public class GridView : Control
{
    private Panel Panel;

    public Frame Header { get; private set; }
    private Frame headerContainer;

    public event SelectedItemChangedEventHandler SelectedItemChanged;

    public GridViewItemCollection Items { get; set; }
    public GridViewItem SelectedItem { get; set; }

    public ActiveList<GridColumn> Columns { get; set; }

    public bool Multiselect { get; set; }

    protected override void Initialize()
    {
        Size = new Point(100, 100);

        Columns = new ActiveList<GridColumn>();
        Columns.ItemAdded += Columns_OnAdd;
        Columns.ItemRemoved += Columns_OnRemove;
        Columns.BeforeItemRemoved += Columns_BeforeItemRemoved;
        Columns.BeforeItemAdded += Columns_BeforeItemAdded;
        Columns.BeforeItemsCleared += Columns_BeforeItemsCleared;

        Items = new GridViewItemCollection();
        Items.ItemAdded += Items_OnAdd;
        Items.ItemRemoved += Items_OnRemove;
        Items.BeforeItemsCleared += Items_BeforeItemsCleared;

        Header = new Frame();
        Header.Size = new Point(20, 32);
        Header.Dock = DockStyle.Top;
        Elements.Add(Header);

        headerContainer = new Frame();
        headerContainer.Size = new Point(20, 32);
        headerContainer.AutoSize = Squid.AutoSize.Horizontal;
        Header.Controls.Add(headerContainer);

        Panel = new Panel();
        Panel.Dock = DockStyle.Fill;
        Panel.Content.AutoSize = Squid.AutoSize.HorizontalVertical;
        //Panel.Margin = new Margin(0, 4, 0, 0);

        Panel.VScroll.ButtonUp.Visible = false;
        Panel.VScroll.ButtonDown.Visible = false;
        Panel.VScroll.Size = new Point(13, 12);
        Panel.VScroll.Slider.Style = "vscrollTrack";
        Panel.VScroll.Slider.Button.Style = "vscrollButton";
        Panel.VScroll.Dock = DockStyle.Right;
        Panel.VScroll.Margin = new Margin(4, 0, 0, 0);

        Panel.HScroll.ButtonUp.Visible = false;
        Panel.HScroll.ButtonDown.Visible = false;
        Panel.HScroll.Size = new Point(13, 12);
        Panel.HScroll.Slider.Style = "vscrollTrack";
        Panel.HScroll.Slider.Button.Style = "vscrollButton";
        Panel.HScroll.Margin = new Margin(0, 4, 0, 0);
        Panel.Content.OnUpdate += Content_OnControlUpdate;
        Elements.Add(Panel);
    }

    protected override void Update()
    {
        base.Update();

        // we want to perform mouse scrolling if:
        // the mouse is scrolling and there is any control hovered
        if (GuiHost.MouseScroll != 0 && Desktop.HotControl != null)
        {
            // ok, lets check if the mouse is anywhere near us
            if (Hit(GuiHost.MousePosition.x, GuiHost.MousePosition.y))
            {
                // now lets check if its really this window or anything in it
                if (Desktop.HotControl == this || Desktop.HotControl.IsChildOf(this))
                    Panel.VScroll.Scroll(GuiHost.MouseScroll);
            }
        }
    }

    void Content_OnControlUpdate(Control sender, SquidEventArgs args)
    {
        headerContainer.Position = new Point(Panel.Content.Position.x, 0);
        headerContainer.PerformLayout();
    }

    void Columns_BeforeItemsCleared(object sender, EventArgs e)
    {
        foreach (GridViewItem item in Items)
            RemoveItem(item);

        foreach (GridColumn col in Columns)
        {
            headerContainer.Controls.Remove(col.Header);
            Panel.Content.Controls.Remove(col.Container);

            col.Container = null;
            col.Header = null;
        }
    }

    void Columns_BeforeItemAdded(object sender, ListEventArgs<GridColumn> e)
    {
        foreach (GridViewItem item in Items)
            RemoveItem(item);
    }

    void Columns_BeforeItemRemoved(object sender, ListEventArgs<GridColumn> e)
    {
        foreach (GridViewItem item in Items)
            RemoveItem(item);
    }

    void Items_BeforeItemsCleared(object sender, EventArgs e)
    {
        foreach (GridViewItem item in Items)
            RemoveItem(item);
    }

    void Columns_OnRemove(object sender, ListEventArgs<GridColumn> e)
    {
        headerContainer.Controls.Remove(e.Item.Header);
        Panel.Content.Controls.Remove(e.Item.Container);

        e.Item.Container = null;
        e.Item.Header = null;

        foreach (GridViewItem item in Items)
            AddItem(item);
    }

    void Columns_OnAdd(object sender, ListEventArgs<GridColumn> e)
    {
        headerContainer.Controls.Add(e.Item.Header);
        Panel.Content.Controls.Add(e.Item.Container);

        foreach (GridViewItem item in Items)
            AddItem(item);
    }

    private void AddItem(GridViewItem item)
    {
        item.Tags.Clear();

        int i = 0;
        foreach (GridColumn column in Columns)
        {
            if (i == 0)
            {
                column.Container.Controls.Add(item);
            }
            else
            {
                int index = i - 1;
                if (item.SubItems.Count > index)
                {
                    column.Container.Controls.Add(item.SubItems[index]);
                    item.Tags.Add(item.SubItems[index]);
                }
                else
                {
                    Control spacer = CreateSpace(item.Size.y);

                    column.Container.Controls.Add(spacer);
                    item.Tags.Add(spacer);
                }
            }

            i++;
        }
    }

    private void RemoveItem(GridViewItem item)
    {
        int i = 0;
        foreach (GridColumn column in Columns)
        {
            if (i == 0)
            {
                column.Container.Controls.Remove(item);
            }
            else
            {
                int index = i - 1;
                if (item.Tags.Count > index)
                {
                    column.Container.Controls.Remove(item.Tags[index]);
                }
            }

            i++;
        }

        item.Tags.Clear();
    }

    private Control CreateSpace(int height)
    {
        Control spacer = new Control();
        spacer.Size = new Point(height, height);
        spacer.Dock = DockStyle.Top;
        return spacer;
    }

    void Items_OnRemove(object sender, ListEventArgs<GridViewItem> e)
    {
        RemoveItem(e.Item);
    }

    void Items_OnAdd(object sender, ListEventArgs<GridViewItem> e)
    {
        AddItem(e.Item);
    }

    void item_OnMouseClick(Control sender, MouseEventArgs args)
    {
        if (args.Button > 0) return;

        GridViewItem item = sender as GridViewItem;

        if (Multiselect)
        {
            item.Selected = !item.Selected;
        }
        else
        {
            if (SelectedItem != null) SelectedItem.Selected = false;
            SelectedItem = item;
            if (SelectedItem != null) SelectedItem.Selected = true;

            // if (SelectedItemChanged != null)
            //     SelectedItemChanged(this, SelectedItem);
        }
    }
}

internal sealed class GridColumnFrame : Frame
{
    internal GridColumnHeader Header;
    private Button ResizeHandle;
    private Point ClickedPos;
    private Point OldSize;

    protected override void Initialize()
    {
        Scissor = true;
        AutoSize = AutoSize.Vertical;
        Dock = DockStyle.Left;

        ResizeHandle = new Button();
        ResizeHandle.Size = new Point(4, 2);
        ResizeHandle.Dock = DockStyle.Right;
        ResizeHandle.OnMouseDown += ResizeHandle_OnMouseDown;
        ResizeHandle.OnMousePress += ResizeHandle_OnMousePress;
        ResizeHandle.Cursor = Cursors.VSplit;
        ResizeHandle.Style = "button";
        Elements.Add(ResizeHandle);
    }

    void ResizeHandle_OnMouseDown(Control sender, MouseEventArgs args)
    {
        if (args.Button > 0) return;

        ClickedPos = GuiHost.MousePosition;
        OldSize = Size;
    }

    void ResizeHandle_OnMousePress(Control sender, MouseEventArgs args)
    {
        if (args.Button > 0) return;

        Point p = GuiHost.MousePosition - ClickedPos;
        ResizeTo(OldSize + p, AnchorStyles.Right);
        Header.Size = new Point(Size.x, Header.Size.y);
    }
}

internal sealed class GridColumnHeader : Label
{
    internal GridColumn Column;

    private Button ResizeHandle;
    private Point ClickedPos;
    private Point OldSize;

    protected override void Initialize()
    {
        base.Initialize();

        Scissor = true;
        Dock = DockStyle.Left;
        Style = "category";

        ResizeHandle = new Button();
        ResizeHandle.Size = new Point(4, 4);
        ResizeHandle.Dock = DockStyle.Right;
        ResizeHandle.OnMouseDown += ResizeHandle_OnMouseDown;
        ResizeHandle.OnMousePress += ResizeHandle_OnMousePress;
        ResizeHandle.Cursor = Cursors.VSplit;
        ResizeHandle.Style = "button";
        Elements.Add(ResizeHandle);
    }

    void ResizeHandle_OnMouseDown(Control sender, MouseEventArgs args)
    {
        if (args.Button > 0) return;

        ClickedPos = GuiHost.MousePosition;
        OldSize = Size;
    }

    void ResizeHandle_OnMousePress(Control sender, MouseEventArgs args)
    {
        if (args.Button > 0) return;

        Point p = GuiHost.MousePosition - ClickedPos;
        ResizeTo(OldSize + p, AnchorStyles.Right);
        Column.Width = Size.x;
    }
}

public class GridColumn
{
    private int _width = 30;
    private string _text = "Column";
    private GridColumnHeader _header;

    public GridColumn()
    {
        Container = new GridColumnFrame();
        Header = new GridColumnHeader();
        Header.Style = "category";

        Container.Header = Header;
        Header.Column = this;
    }

    internal GridColumnFrame Container;

    internal GridColumnHeader Header
    {
        get { return _header; }
        set
        {
            _header = value;

            if (_header != null)
                _header.Text = Text;
        }
    }

    public int Width
    {
        get
        {
            return _width;
        }
        set
        {
            if (_width == value)
                return;

            _width = value;

            if (Header != null)
                Header.Size = new Point(_width, Header.Size.y);

            if (Container != null)
                Container.Size = new Point(_width, Container.Size.y);
        }
    }

    public string Text
    {
        get { return _text; }
        set
        {
            _text = value;

            if (Header != null)
                Header.Text = value;
        }
    }
}

public class GridViewItem : GridViewSubItem
{
    internal List<Control> Tags = new List<Control>();

    public object Value { get; set; }

    public ActiveList<GridViewSubItem> SubItems { get; set; }

    protected override void Initialize()
    {
        base.Initialize();

        SubItems = new ActiveList<GridViewSubItem>();
        Size = new Point(100, 20);
        Dock = DockStyle.Top;
    }
}

public class GridViewSubItem : Label
{
    public event SelectedChangedEventHandler OnSelectedChanged;

    private bool _selected;
    [DefaultValue(false)]
    public new bool Selected
    {
        get { return _selected; }
        set
        {
            if (value == _selected) return;

            _selected = value;

            if (OnSelectedChanged != null)
                OnSelectedChanged(this);
        }
    }
    protected override void Initialize()
    {
        Size = new Point(100, 20);
        Dock = DockStyle.Top;
    }
}
